@@ -1,6 +1,5 @@
-from typing import Dict, Optional, List, Union
+from typing import Dict, Optional, List, Union, Any
 import socket
-import logging
 import pdb
 import pathlib
 import os
@@ -8,23 +7,31 @@ import time
 import datetime
 import json
 import random
+import tempfile
+import zipfile
+from concurrent.futures import wait
 
+# Third-party Imports
 import dill
-
-logger = logging.getLogger("chimerapy")
-
+import aiohttp
+from aiohttp import web
 import networkx as nx
 
-from .server import Server
+from .networking import Server, Client
+from .networking.enums import MANAGER_MESSAGE, WORKER_MESSAGE
 from .graph import Graph
-from . import enums
+from .exceptions import CommitGraphError
+from . import _logger
+from .utils import waiting_for
+
+logger = _logger.getLogger("chimerapy")
 
 
 class Manager:
     def __init__(
         self,
         logdir: pathlib.Path,
-        port: int = 9000,
+        port: int = 8080,
         max_num_of_workers: int = 50,
     ):
         """Create ``Manager``, the controller of the cluster.
@@ -40,61 +47,70 @@ class Manager:
 
         """
         # Saving input parameters
-        # self.host = socket.gethostbyname(socket.gethostname())
+        self.name = "Manager"
+        self.host = "localhost"
         self.port = port
         self.max_num_of_workers = max_num_of_workers
+        self.has_shutdown = False
 
         # Create log directory to store data
-        timestamp = datetime.datetime.now().strftime("%Y_%m_%d:%H_%M_%S")
+        timestamp = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
         rand_num = random.randint(1000, 9999)
         self.logdir = logdir / f"chimerapy-{timestamp}-{rand_num}"
 
+        # Also create a tempfolder to store any miscellaneous files and folders
+        self.tempfolder = pathlib.Path(tempfile.mkdtemp())
+
         # Create a logging directory
         os.makedirs(self.logdir, exist_ok=True)
 
         # Instance variables
-        self.workers: Dict = {}
+        self.workers: Dict[str, Dict[str, Any]] = {}
         self.graph: Graph = Graph()
         self.worker_graph_map: Dict = {}
         self.commitable_graph: bool = False
         self.nodes_server_table: Dict = {}
         self.start_time: Optional[datetime.datetime] = None
         self.stop_time: Optional[datetime.datetime] = None
-
-        # Specifying the handlers for worker2manager communication
-        self.handlers = {
-            enums.WORKER_REGISTER: self.register_worker,
-            enums.WORKER_DEREGISTER: self.deregister_worker,
-            enums.WORKER_REPORT_NODE_SERVER_DATA: self.node_server_data,
-            enums.WORKER_REPORT_NODES_STATUS: self.update_nodes_status,
-            enums.WORKER_COMPLETE_BROADCAST: self.complete_worker_broadcast,
-            enums.WORKER_REPORT_GATHER: self.get_gather,
-            enums.WORKER_TRANSFER_COMPLETE: self.complete_worker_transfer,
-        }
+        self.duration: int = 0
 
         # Create server
         self.server = Server(
             port=self.port,
             name="Manager",
-            max_num_of_clients=self.max_num_of_workers,
-            sender_msg_type=enums.MANAGER_MESSAGE,
-            accepted_msg_type=enums.WORKER_MESSAGE,
-            handlers=self.handlers,
+            ws_handlers={
+                WORKER_MESSAGE.REGISTER: self.register_worker,
+                WORKER_MESSAGE.DEREGISTER: self.deregister_worker,
+                WORKER_MESSAGE.REPORT_NODE_SERVER_DATA: self.node_server_data,
+                WORKER_MESSAGE.REPORT_NODES_STATUS: self.update_nodes_status,
+                WORKER_MESSAGE.COMPLETE_BROADCAST: self.complete_worker_broadcast,
+                WORKER_MESSAGE.REPORT_GATHER: self.get_gather,
+                WORKER_MESSAGE.PACKAGE_LOADED: self.complete_worker_load_package,
+                WORKER_MESSAGE.TRANSFER_COMPLETE: self.complete_worker_transfer,
+            },
         )
-        self.server.start()
+        self.server.serve()
         logger.info(f"Manager started at {self.server.host}:{self.server.port}")
 
         # Updating the manager's port to the found available port
         self.host, self.port = self.server.host, self.server.port
 
+    def __repr__(self):
+        return f"<Manager @{self.host}:{self.port}>"
+
+    def __str__(self):
+        return self.__repr__()
+
     ####################################################################
     ## Message Reactivity API
     ####################################################################
 
-    def register_worker(self, msg: Dict, worker_socket: socket.socket):
+    async def register_worker(self, msg: Dict, ws: web.WebSocketResponse):
         self.workers[msg["data"]["name"]] = {
             "addr": msg["data"]["addr"],
-            "socket": worker_socket,
+            "http_port": msg["data"]["http_port"],
+            "http_ip": msg["data"]["http_ip"],
+            "ws": ws,
             "ack": True,
             "response": False,
             "reported_nodes_server_data": False,
@@ -102,19 +118,20 @@ class Manager:
             "nodes_status": {},
             "gather": {},
             "collection_complete": False,
+            "package_loaded": False,
         }
         logger.info(
             f"Manager registered <Worker name={msg['data']['name']}> from {msg['data']['addr']}"
         )
 
-    def deregister_worker(self, msg: Dict, worker_socket: socket.socket):
-        worker_socket.close()
+    async def deregister_worker(self, msg: Dict, ws: web.WebSocketResponse):
+        await ws.close()
         logger.info(
             f"Manager deregistered <Worker name={msg['data']['name']}> from {msg['data']['addr']}"
         )
         del self.workers[msg["data"]["name"]]
 
-    def node_server_data(self, msg: Dict, worker_socket: socket.socket):
+    async def node_server_data(self, msg: Dict, ws: web.WebSocketResponse):
 
         # And then updating the node server data
         self.nodes_server_table.update(msg["data"]["nodes"])
@@ -122,7 +139,7 @@ class Manager:
         # Tracking which workers have responded
         self.workers[msg["data"]["name"]]["reported_nodes_server_data"] = True
 
-    def update_nodes_status(self, msg: Dict, worker_socket: socket.socket):
+    async def update_nodes_status(self, msg: Dict, ws: web.WebSocketResponse):
 
         # Updating nodes status
         self.workers[msg["data"]["name"]]["nodes_status"] = msg["data"]["nodes_status"]
@@ -130,17 +147,22 @@ class Manager:
 
         logger.debug(f"{self}: Nodes status update to: {self.workers}")
 
-    def complete_worker_broadcast(self, msg: Dict, worker_socket: socket.socket):
+    async def complete_worker_broadcast(self, msg: Dict, ws: web.WebSocketResponse):
 
         # Tracking which workers have responded
         self.workers[msg["data"]["name"]]["served_nodes_server_data"] = True
 
-    def get_gather(self, msg: Dict, worker_socket: socket.socket):
+    async def get_gather(self, msg: Dict, ws: web.WebSocketResponse):
 
         self.workers[msg["data"]["name"]]["gather"] = msg["data"]["node_data"]
         self.workers[msg["data"]["name"]]["response"] = True
 
-    def complete_worker_transfer(self, msg: Dict, worker_socket: socket.socket):
+    async def complete_worker_load_package(self, msg: Dict, ws: web.WebSocketResponse):
+
+        self.workers[msg["data"]["name"]]["package_loaded"] = True
+        self.workers[msg["data"]["name"]]["response"] = True
+
+    async def complete_worker_transfer(self, msg: Dict, ws: web.WebSocketResponse):
 
         self.workers[msg["data"]["name"]]["collection_complete"] = True
         self.workers[msg["data"]["name"]]["response"] = True
@@ -222,19 +244,17 @@ class Manager:
 
         # Send for the creation of the node
         self.server.send(
-            self.workers[worker_name]["socket"],
-            {
-                "signal": enums.MANAGER_CREATE_NODE,
-                "data": {
-                    "worker_name": worker_name,
-                    "node_name": node_name,
-                    "pickled": dill.dumps(
-                        self.graph.G.nodes[node_name]["object"], recurse=True
-                    ),
-                    "in_bound": list(self.graph.G.predecessors(node_name)),
-                    "out_bound": list(self.graph.G.successors(node_name)),
-                    "follow": self.graph.G.nodes[node_name]["follow"],
-                },
+            client_name=worker_name,
+            signal=MANAGER_MESSAGE.CREATE_NODE,
+            data={
+                "worker_name": worker_name,
+                "node_name": node_name,
+                "pickled": dill.dumps(
+                    self.graph.G.nodes[node_name]["object"], recurse=True
+                ),
+                "in_bound": list(self.graph.G.predecessors(node_name)),
+                "out_bound": list(self.graph.G.successors(node_name)),
+                "follow": self.graph.G.nodes[node_name]["follow"],
             },
         )
 
@@ -278,8 +298,9 @@ class Manager:
 
                 # Send the request to each worker
                 self.server.send(
-                    self.workers[worker_name]["socket"],
-                    {"signal": enums.MANAGER_REQUEST_NODE_SERVER_DATA, "data": {}},
+                    client_name=worker_name,
+                    signal=MANAGER_MESSAGE.REQUEST_NODE_SERVER_DATA,
+                    data={},
                 )
 
                 # Wait until response
@@ -315,11 +336,9 @@ class Manager:
 
                 # Send the request to each worker
                 self.server.send(
-                    self.workers[worker_name]["socket"],
-                    {
-                        "signal": enums.MANAGER_BROADCAST_NODE_SERVER_DATA,
-                        "data": self.nodes_server_table,
-                    },
+                    client_name=worker_name,
+                    signal=MANAGER_MESSAGE.BROADCAST_NODE_SERVER_DATA,
+                    data=self.nodes_server_table,
                 )
 
                 # Wait until response
@@ -412,6 +431,67 @@ class Manager:
 
             miss_counter += 1
 
+    def register_graph(self, graph: Graph):
+        """Verifying that a Graph is valid, that is a DAG.
+
+        In ChimeraPy, cycle are not allowed, this is to avoid a deadlock.
+        Registering the graph is the first step to setting up the data
+        pipeline in  the cluster.
+
+        Args:
+            graph (Graph): A directed acyclic graph.
+
+        """
+        # First, check if graph is valid
+        if not graph.is_valid():
+            logger.error("Invalid Graph - rejected!")
+            raise CommitGraphError("Invalid Graph, not DAG - rejected!")
+
+        # Else, let's save it
+        self.graph = graph
+
+    def map_graph(self, worker_graph_map: Dict[str, List[str]]):
+        """Mapping ``Node`` from graph to ``Worker`` from cluster.
+
+        The mapping, a dictionary, informs ChimeraPy which ``Worker`` is
+        going to execute which ``Node``s.
+
+        Args:
+            worker_graph_map (Dict[str, List[str]]): The keys are the \
+            ``Worker``'s name and the values should be a list of \
+            the ``Node``'s names that will be executed within its corresponding\
+            ``Worker`` key.
+
+        """
+        # Tracking valid inputs
+        checks = []
+
+        # First, check if the mapping is valid!
+        for worker_name in worker_graph_map:
+
+            # First check if the worker is registered!
+            if worker_name not in self.workers:
+                logger.error(f"{worker_name} is not register to Manager.")
+                checks.append(False)
+                break
+            else:
+                checks.append(True)
+
+            # Then check if the node exists in the graph
+            for node_name in worker_graph_map[worker_name]:
+                if not self.graph.has_node_by_name(node_name):
+                    logger.error(f"{node_name} is not in Manager's graph.")
+                    checks.append(False)
+                    break
+                else:
+                    checks.append(True)
+
+        if not all(checks):
+            raise CommitGraphError("Mapping is invalid")
+
+        # Save the worker graph
+        self.worker_graph_map = worker_graph_map
+
     ####################################################################
     ## Cluster Setup, Control, and Monitor API
     ####################################################################
@@ -461,76 +541,58 @@ class Manager:
         # Distribute the entire graph's information to all the Workers
         self.request_connection_creation()
 
+    ####################################################################
+    ## Package and Dependency Management
+    ####################################################################
+
+    def distribute_packages(self, packages_meta: List[Dict[str, Any]]):
+
+        # For each package requested, send the packages to be used
+        # by the connected Workers.
+        for package_meta in packages_meta:
+            package_name = package_meta["name"]
+            package_path = package_meta["path"]
+
+            # We do so by sending compressed zip files of the packages
+            zip_package_dst = self.tempfolder / f"{package_name}.zip"
+            with zipfile.PyZipFile(str(zip_package_dst), mode="w") as zip_pkg:
+                zip_pkg.writepy(package_path)
+
+            # Send it to the workers and let them know to load the
+            # send package
+            for worker_name in self.workers:
+
+                worker_data = self.workers[worker_name]
+
+                # Create a temporary HTTP client
+                client = Client(
+                    self.name,
+                    host=worker_data["http_ip"],
+                    port=worker_data["http_port"],
+                )
+                client.send_file(sender_name=self.name, filepath=zip_package_dst)
+
+        # Send package finish confirmation
+        self.server.broadcast(
+            signal=MANAGER_MESSAGE.REQUEST_CODE_LOAD,
+            data={"packages": [x["name"] for x in packages_meta]},
+        )
+
+        # Wait until workers have obtain the files and loaded to the path
+        # TODO
+        # Wail until all workers have responded with their node server data
+        self.wait_until_all_workers_responded(attribute="package_loaded", timeout=30)
+
     ####################################################################
     ## User API
     ####################################################################
 
-    def register_graph(self, graph: Graph):
-        """Verifying that a Graph is valid, that is a DAG.
-
-        In ChimeraPy, cycle are not allowed, this is to avoid a deadlock.
-        Registering the graph is the first step to setting up the data
-        pipeline in  the cluster.
-
-        Args:
-            graph (Graph): A directed acyclic graph.
-
-        """
-        # First, check if graph is valid
-        if not graph.is_valid():
-            logger.error("Invalid Graph - rejected!")
-            raise RuntimeError("Invalid Graph - rejected!")
-
-        # Else, let's save it
-        self.graph = graph
-        self.commitable_graph = False
-
-    def map_graph(self, worker_graph_map: Dict[str, List[str]]):
-        """Mapping ``Node`` from graph to ``Worker`` from cluster.
-
-        The mapping, a dictionary, informs ChimeraPy which ``Worker`` is
-        going to execute which ``Node``s.
-
-        Args:
-            worker_graph_map (Dict[str, List[str]]): The keys are the \
-            ``Worker``'s name and the values should be a list of \
-            the ``Node``'s names that will be executed within its corresponding\
-            ``Worker`` key.
-
-        """
-        # Tracking valid inputs
-        checks = []
-
-        # First, check if the mapping is valid!
-        for worker_name in worker_graph_map:
-
-            # First check if the worker is registered!
-            if worker_name not in self.workers:
-                logger.error(f"{worker_name} is not register to Manager.")
-                checks.append(False)
-                break
-            else:
-                checks.append(True)
-
-            # Then check if the node exists in the graph
-            for node_name in worker_graph_map[worker_name]:
-                if not self.graph.has_node_by_name(node_name):
-                    logger.error(f"{node_name} is not in Manager's graph.")
-                    checks.append(False)
-                    break
-                else:
-                    checks.append(True)
-
-        # If everything is okay, we are approved to commit the graph
-        if all(checks):
-            self.commitable_graph = True
-        else:
-            self.commitable_graph = False
-
-        # Save the worker graph
-        self.worker_graph_map = worker_graph_map
-
-    def commit_graph(self, timeout: Optional[Union[int, float]] = None):
+    def commit_graph(
+        self,
+        graph: Graph,
+        mapping: Dict[str, List[str]],
+        send_packages: Optional[List[Dict[str, Any]]] = None,
+    ):
         """Committing ``Graph`` to the cluster.
 
         Committing refers to how the graph itself (with its nodes and edges)
@@ -549,11 +611,27 @@ class Manager:
         ``Nodes`` are located; thereby establishing the edges between
         ``Nodes``.
 
+        Args:
+            graph (cp.Graph): The graph to deploy within the cluster.
+            mapping (Dict[str, List[str]): Mapping from ``cp.Worker`` to\
+                ``cp.Nodes`` through a dictionary. The keys are the name\
+                of the workers, while the value is a list of the nodes' \
+                names.
+            send_packages (Optional[List[Dict[str, Any]]]): An optional
+                feature for transferring a local package (typically a \
+                development package not found via PYPI or Anaconda). \
+                Provide a list of packages with each package configured \
+                via dictionary with the following key-value pairs: \
+                name:``str`` and path:``pathlit.Path``.
+
         """
-        # First, check that the graph has been cleared
-        assert self.commitable_graph, logger.error(
-            "Tried to commit graph that hasn't been confirmed or ready"
-        )
+        # First, test that the graph and the mapping are valid
+        self.register_graph(graph)
+        self.map_graph(mapping)
+
+        # Then send requested packages
+        if send_packages:
+            self.distribute_packages(send_packages)
 
         # First, create the network
         self.create_p2p_network()
@@ -580,7 +658,7 @@ class Manager:
 
         # Send message for workers to inform all nodes to take a single
         # step
-        self.server.broadcast({"signal": enums.MANAGER_REQUEST_STEP, "data": {}})
+        self.server.broadcast(signal=MANAGER_MESSAGE.REQUEST_STEP, data={})
 
     def gather(self) -> Dict:
 
@@ -592,7 +670,7 @@ class Manager:
         self.mark_response_as_false_for_workers()
 
         # Request gathering the latest value of each node
-        self.server.broadcast({"signal": enums.MANAGER_REQUEST_GATHER, "data": {}})
+        self.server.broadcast(signal=MANAGER_MESSAGE.REQUEST_GATHER, data={})
 
         # Wail until all workers have responded with their node server data
         self.wait_until_all_workers_responded(attribute="gather", timeout=5)
@@ -620,7 +698,7 @@ class Manager:
         self.start_time = datetime.datetime.now()
 
         # Tell the cluster to start
-        self.server.broadcast({"signal": enums.MANAGER_START_NODES, "data": {}})
+        self.server.broadcast(signal=MANAGER_MESSAGE.START_NODES, data={})
 
     def stop(self):
         """Stop the executiong of the cluster.
@@ -631,26 +709,32 @@ class Manager:
         """
         # Mark the start time
         self.stop_time = datetime.datetime.now()
+        self.duration = (self.stop_time - self.start_time).total_seconds()
 
         # Tell the cluster to stop
-        self.server.broadcast({"signal": enums.MANAGER_STOP_NODES, "data": {}})
+        self.server.broadcast(signal=MANAGER_MESSAGE.STOP_NODES, data={})
 
     def collect(self, timeout: Optional[Union[int, float]] = None, unzip: bool = True):
         """Collect data archives across the cluster to the Manager's logs."""
         # Wait until the nodes first finished writing down the data
         self.wait_until_all_nodes_saved(timeout=timeout)
 
-        # Request for archives to be send
-        self.server.broadcast({"signal": enums.MANAGER_REQUEST_COLLECT, "data": {}})
-
-        # Wail until all workers have responded with their node server data
-        self.wait_until_all_workers_responded(
-            attribute="collection_complete", timeout=timeout
+        # Request
+        self.server.broadcast(
+            signal=MANAGER_MESSAGE.REQUEST_COLLECT, data={"path": str(self.logdir)}
         )
 
+        # Collect from each worker individually
+        for worker_name in self.workers:
+
+            # Wait until
+            self.wait_until_worker_respond(
+                worker_name, attribute="collection_complete", timeout=self.duration
+            )
+
         # # Then move the tempfiles to the log runs and unzip
         self.server.move_transfer_files(self.logdir, unzip)
-        logger.info(f"{self}: data collection complete!")
+        logger.info(f"{self}: Data collection complete!")
 
         # Add the meta data to the archive
         self.save_meta()
@@ -662,5 +746,36 @@ class Manager:
         to shutdown, in which they will stop their processes and threads safely.
 
         """
+
+        # Only let shutdown happen once
+        if self.has_shutdown:
+            logger.debug(f"{self}: requested to shutdown twice, skipping.")
+            return None
+        else:
+            self.has_shutdown = True
+
+        # If workers are connected, let's notify them that the cluster is
+        # shutting down
+        if len(self.workers) > 0:
+
+            # Send shutdown message
+            self.server.broadcast(signal=MANAGER_MESSAGE.CLUSTER_SHUTDOWN, data={})
+
+            # Wait until all worker's deregister
+            waiting_for(
+                condition=lambda: len(self.workers) == 0,
+                check_period=0.1,
+                success_msg=f"{self}: All workers shutdown",
+                timeout=10,
+                timeout_raise=False,
+                timeout_msg=f"{self}: Not all workers shutdown -> {self.workers}",
+            )
+
         # First, shutdown server
         self.server.shutdown()
+
+    def __del__(self):
+
+        # Also good to shutdown anything that isn't
+        if not self.has_shutdown:
+            self.shutdown()
