@@ -1,123 +1,24 @@
-from typing import Any, Dict, Tuple
-import socket
+from typing import Callable, Union, Optional, Any, Dict
 import queue
-import errno
-import datetime
-import threading
 import logging
 import functools
-import struct
+import time
+import enum
 import pickle
+import datetime
+import uuid
+import socket
+import asyncio
 
-import netifaces as ni
+# Third-party
 from tqdm import tqdm
-import lz4.block
+import blosc
+import netifaces as ni
 
-logger = logging.getLogger("chimerapy")
+# Internal
+from . import _logger
 
-
-def threaded(fn):
-    """Decorator for class methods to be spawn new thread.
-
-    From: https://stackoverflow.com/a/19846691/13231446
-    Args:
-        fn: The method of a class.
-    """
-
-    def wrapper(*args, **kwargs):
-        thread = threading.Thread(target=fn, args=args, kwargs=kwargs)
-        thread.deamon = True
-        return thread
-
-    return wrapper
-
-
-def log(func):
-    @functools.wraps(func)
-    def wrapper(*args, **kwargs):
-        args_repr = [repr(a) for a in args]
-        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
-        signature = ", ".join(args_repr + kwargs_repr)
-        # logger.debug(f"function {func.__name__} called with args {signature}")
-        logger.debug(f"{args_repr[0]}: function {func.__name__}")
-        try:
-            result = func(*args, **kwargs)
-            return result
-        except Exception as e:
-            logger.exception(
-                f"Exception raised in {func.__name__}. exception: {str(e)}"
-            )
-            raise e
-
-    return wrapper
-
-
-def get_open_port(start_port: int) -> socket.socket:
-
-    # Creating socket to connect
-    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-    offset = 0
-
-    while True:
-        try:
-            current_attempt_port = start_port + offset
-            s.bind(("", current_attempt_port))
-            break
-        except socket.error as e:
-            offset += 10
-            if e.errno == errno.EADDRINUSE:
-                logger.debug(f"Port {current_attempt_port} is already in use.")
-            else:
-                logger.error("Unknown socket error", exc_info=True)
-                raise e
-
-    return s
-
-
-def get_ip_address() -> str:
-
-    # Get gateway of the network
-    gws = ni.gateways()
-    default_gw_name = gws["default"][ni.AF_INET][1]
-
-    # Get the ip in the default gateway
-    ip = ni.ifaddresses(default_gw_name)[ni.AF_INET][0]["addr"]
-    return ip
-
-
-def create_payload(
-    data: Any,
-    type: str,
-    signal: str,
-    provided_uuid: str,
-    ack: bool,
-    timestamp: datetime.timedelta = datetime.timedelta(),
-) -> Tuple[bytes, bytes]:
-
-    payload = {
-        "type": type,
-        "signal": signal,
-        "timestamp": str(timestamp),
-        "uuid": provided_uuid,
-        "data": data,
-        "ack": int(ack),
-    }
-
-    b_payload = pickle.dumps(payload)
-    compressed_bytes_payload = lz4.block.compress(b_payload)
-
-    finished_payload = compressed_bytes_payload
-
-    return finished_payload, struct.pack(">Q", len(finished_payload))
-
-
-def decode_payload(data: bytes) -> Dict:
-
-    bytes_payload = lz4.block.decompress(data)
-    # bytes_payload = gzip.decompress(data)
-    payload: Dict = pickle.loads(bytes_payload)
-
-    return payload
+logger = _logger.getLogger("chimerapy")
 
 
 def clear_queue(input_queue: queue.Queue):
@@ -173,3 +74,117 @@ class logging_tqdm(tqdm):
         if not msg:
             msg = self.__str__()
         self.logger.info("%s", msg)
+
+
+async def async_waiting_for(
+    condition: Callable[[], bool],
+    check_period: Union[int, float] = 0.1,
+    success_msg: Optional[str] = None,
+    timeout: Optional[Union[int, float]] = None,
+    timeout_raise: Optional[bool] = True,
+    timeout_msg: Optional[str] = None,
+) -> bool:
+
+    counter = 0
+    while True:
+
+        if condition():
+            if success_msg:
+                logger.debug(success_msg)
+            return True
+        else:
+            await asyncio.sleep(check_period)
+            counter += 1
+
+            if timeout and counter * check_period > timeout:
+                if timeout_raise:
+                    raise TimeoutError(timeout_msg)
+                else:
+                    if timeout_msg:
+                        logger.debug(timeout_msg)
+                    return False
+
+
+def waiting_for(
+    condition: Callable[[], bool],
+    check_period: Union[int, float] = 0.1,
+    success_msg: Optional[str] = None,
+    timeout: Optional[Union[int, float]] = None,
+    timeout_raise: Optional[bool] = True,
+    timeout_msg: Optional[str] = None,
+) -> bool:
+
+    counter = 0
+    while True:
+
+        if condition():
+            if success_msg:
+                logger.debug(success_msg)
+            return True
+        else:
+            time.sleep(check_period)
+            counter += 1
+
+            if timeout and counter * check_period > timeout:
+                if timeout_raise:
+                    raise TimeoutError(timeout_msg)
+                else:
+                    if timeout_msg:
+                        logger.debug(timeout_msg)
+                    return False
+
+
+def get_open_port(start_port: int) -> socket.socket:
+
+    # Creating socket to connect
+    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+    offset = 0
+
+    while True:
+        try:
+            current_attempt_port = start_port + offset
+            s.bind(("", current_attempt_port))
+            break
+        except socket.error as e:
+            offset += 10
+            if e.errno == errno.EADDRINUSE:
+                logger.debug(f"Port {current_attempt_port} is already in use.")
+            else:
+                logger.error("Unknown socket error", exc_info=True)
+                raise e
+
+    return s
+
+
+def get_ip_address() -> str:
+
+    # Get gateway of the network
+    gws = ni.gateways()
+    default_gw_name = gws["default"][ni.AF_INET][1]
+
+    # Get the ip in the default gateway
+    ip = ni.ifaddresses(default_gw_name)[ni.AF_INET][0]["addr"]
+    return ip
+
+
+def create_payload(
+    signal: enum.Enum,
+    data: Any,
+    msg_uuid: str = str(uuid.uuid4()),
+    timestamp: datetime.timedelta = datetime.timedelta(),
+    ok: bool = False,
+) -> bytes:
+
+    payload = {
+        "signal": signal,
+        "timestamp": str(timestamp),
+        "data": data,
+        "uuid": msg_uuid,
+        "ok": ok,
+    }
+
+    return blosc.compress(pickle.dumps(payload, protocol=pickle.HIGHEST_PROTOCOL))
+
+
+def decode_payload(data: bytes) -> Dict[str, Any]:
+    return pickle.loads(blosc.decompress(data))
