@@ -5,12 +5,13 @@ import sys
 
 # Third-party Imports
 import pytest
+import chimerapy as cp
 
 # Test Import
 from .mock import DockeredWorker
 from .conftest import linux_run_only
 
-logger = logging.getLogger("chimerapy")
+logger = cp._logger.getLogger("chimerapy")
 
 
 @linux_run_only
@@ -38,6 +39,7 @@ def test_multiple_workers_connect(manager, docker_client):
     for worker in workers:
         assert worker.name in manager.workers
 
+    logger.info("Manager shutting down")
     manager.shutdown()
 
     time.sleep(1)
