@@ -9,7 +9,9 @@ from pytest_lazyfixture import lazy_fixture
 
 import chimerapy as cp
 
-logger = logging.getLogger("chimerapy")
+cp.debug()
+
+logger = cp._logger.getLogger("chimerapy")
 
 
 class LowFrequencyNode(cp.Node):
@@ -17,14 +19,17 @@ class LowFrequencyNode(cp.Node):
         self.i = 0
 
     def step(self):
+        data_chunk = cp.DataChunk()
         if self.i == 0:
             time.sleep(0.5)
             self.i += 1
-            return self.i
+            data_chunk.add("i", self.i)
+            return data_chunk
         else:
             time.sleep(3)
             self.i += 1
-            return self.i
+            data_chunk.add("i", self.i)
+            return data_chunk
 
 
 class HighFrequencyNode(cp.Node):
@@ -34,19 +39,24 @@ class HighFrequencyNode(cp.Node):
     def step(self):
         time.sleep(0.1)
         self.i += 1
-        return self.i
+        data_chunk = cp.DataChunk()
+        data_chunk.add("i", self.i)
+        return data_chunk
 
 
 class SubsequentNode(cp.Node):
     def prep(self):
         self.record = {}
 
-    def step(self, data: Dict[str, Any]):
+    def step(self, data: Dict[str, cp.DataChunk]):
 
         for k, v in data.items():
             self.record[k] = v
 
-        return self.record
+        data_chunk = cp.DataChunk()
+        data_chunk.add("record", self.record)
+
+        return data_chunk
 
 
 @pytest.fixture
@@ -80,53 +90,53 @@ def step_down_graph():
 
 
 @pytest.mark.parametrize(
-    "_manager, _worker, config_graph, follow",
+    "config_graph, follow",
     [
         (
-            lazy_fixture("manager"),
-            lazy_fixture("worker"),
             lazy_fixture("step_up_graph"),
             "up",
         ),
-        (
-            lazy_fixture("manager"),
-            lazy_fixture("worker"),
-            lazy_fixture("step_down_graph"),
-            "down",
-        ),
+        # (
+        #     lazy_fixture("step_down_graph"),
+        #     "down",
+        # ),
     ],
 )
-def test_node_frequency_execution(_manager, _worker, config_graph, follow):
+def test_node_frequency_execution(manager, worker, config_graph, follow):
 
-    # Connect to the _manager
-    _worker.connect(host=_manager.host, port=_manager.port)
+    # Connect to the manager
+    worker.connect(host=manager.host, port=manager.port)
 
-    # Then register graph to _manager
-    _manager.register_graph(config_graph)
-
-    # Specify what nodes to what _worker
-    _manager.map_graph(
+    # Then register graph to manager
+    manager.commit_graph(
+        config_graph,
         {
             "local": ["hf", "lf", "sn"],
-        }
+        },
     )
 
-    # Commiting the graph by sending it to the _workers
-    _manager.commit_graph(timeout=10)
-
     # Take a single step and see if the system crashes and burns
-    _manager.start()
+    manager.start()
     time.sleep(3)
-    _manager.stop()
+    manager.stop()
 
     # Then request gather and confirm that the data is valid
-    latest_data_values = _manager.gather()
+    latest_data_values = manager.gather()
     logger.info(f"Data Values: {latest_data_values}")
-    step_records = latest_data_values["sn"]
+    data_chunk_step_records = latest_data_values["sn"]
+    step_records = data_chunk_step_records.get("record")["value"]
 
     if follow == "up":
         assert step_records["lf"] == latest_data_values["lf"]
-        assert np.abs((step_records["hf"] - latest_data_values["hf"])) < 5
+        assert (
+            np.abs(
+                (
+                    step_records["hf"].get("i")["value"]
+                    - latest_data_values["hf"].get("i")["value"]
+                )
+            )
+            < 5
+        )
     elif follow == "down":
         assert step_records["lf"] == latest_data_values["lf"]
         assert step_records["hf"] != latest_data_values["hf"]
